@@ -264,9 +264,19 @@ def s3downloader(**kwargs):
     glb = kwargs.pop('globals')
     tmpfile_dir = kwargs.pop('tmpfile_dir')
 
-    # 1. tmpfile_dir=None, glb=None: Read object into memory and return bytes to pipe
-    # 2. tmpfile_dir is present and glb=None: Read object into tmpfile_dir and return status
-    # 3. tmpfile_dir is present and glb is present: Read object into tmpfile_dir and call
+    # num_threads() processes are forked. Each child process executes this function
+    # This function reads operations(op) from the parent process through a Pipe.
+    # The only ops we support today are download and quit
+    # For download, we have three distnct modes we operate in:
+    # Mode 1 is download into memory and write the bytes to the parent process through the Pipe
+    # Mode 2 is for downloading to a temporary dir. Returns status to parent process through Pipe
+    # Mode 3 is where we download one object and execute the function provided by the user,
+    # the xform_one_object function written by the Data Scientist
+    #
+    # Here is how we recognize the different modes this function operates in:
+    # Mode 1. tmpfile_dir=None, glb=None: Read object into memory and return bytes to pipe
+    # Mode 2. tmpfile_dir is present and glb=None: Read object into tmpfile_dir and return status
+    # Mode 3. tmpfile_dir is present and glb is present: Read object into tmpfile_dir and call
     #    the function infin_transform_one_object using glb as globals. Send status back
     # Note: If glb is present, tmpfile_dir must be present. We always call
     # infin_transform_one_object by passing in a temporary file
@@ -340,10 +350,13 @@ def s3downloader(**kwargs):
                     write_pipe.send(key)
                     write_pipe.send(filename)
                     if (tmpfile_dir == None):
+                        write_pipe.send(content_length)
                         for chunk in strbody.iter_chunks(8192):
                             write_pipe.send(chunk)
                     else:
-                        write_pipe.send('Success')
+                        status_as_byte_array = bytes('Success', 'utf-8')
+                        write_pipe.send(len(status_as_byte_array))
+                        write_pipe.send(status_as_byte_array)
                 except Exception as e:
                     status_str = str(e)
                     info("Error sending bytes back: " + status_str + ", traceback = \n" + traceback.format_exc())
@@ -365,7 +378,9 @@ def s3downloader(**kwargs):
                 os.remove(tmpf_name)
                 write_pipe.send(key)
                 write_pipe.send(filename)
-                write_pipe.send(status_str)
+                status_as_byte_array = bytes(status_str, 'utf-8')
+                write_pipe.send(len(status_as_byte_array))
+                write_pipe.send(status_as_byte_array)
         elif (op == 'quit'):
             if (verbose == True):
                 info('s3downloader: Received quit command')
@@ -442,22 +457,37 @@ def download_objects_inner(client, bucketname, parentdir, prefix_trunc, array_of
         files_read = 0
         while pipe_from_child:
             ready = wait(pipe_from_child, timeout=10)
-            for read_pipe in pipe_from_child:
+            for read_pipe in ready:
+                key = None
+                fn = None
+                length = None
+                this_file_bytes = None
                 try:
-                    while (read_pipe.poll()):
-                        key = read_pipe.recv()
-                        fn = read_pipe.recv()
+                    read_pipe.poll(None)
+                    key = read_pipe.recv()
+                    read_pipe.poll(None)
+                    fn = read_pipe.recv()
+                    read_pipe.poll(None)
+                    length = read_pipe.recv()
+                    this_file_bytes = bytearray()
+                    bytes_read = 0
+                    while (read_pipe.poll(None)):
                         bts = read_pipe.recv()
-                        filekeys.append(key)
-                        filenames.append(fn)
-                        filebytes.append(bts)
-                        pbar.update(1)
-                        files_read += 1
-                        if ((files_read % 10) == 0):
-                            log_metric("downloaded", files_read, step=step)
-                            step += 1
+                        bytes_read = bytes_read + len(bts)
+                        this_file_bytes.extend(bts)
+                        if (bytes_read == length):
+                            break
                 except EOFError:
                     pipe_from_child.remove(read_pipe)
+                if (key and fn and this_file_bytes and len(this_file_bytes) > 0):
+                    filekeys.append(key)
+                    filenames.append(fn)
+                    filebytes.append(this_file_bytes)
+                    pbar.update(1)
+                    files_read += 1
+                    if ((files_read % 10) == 0):
+                        log_metric("downloaded", files_read, step=step)
+                        step += 1
 
     for i in range(num_threads()):
         processes[i].join()
